/* $Id: main.c,v 1.20 1998/02/23 02:13:15 mijan Exp $ */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "util.h"
#include "inner.h"
#include "outer_control.h"

#define PHRASE "smuggle munitions KGB NORAD Albanian Serbian Ortega cracking radar"

void innerControlVariate(int whichInnerControl,
			 meshData *mesh_in, meshData *mesh_out);

void sumWeights(meshData *md);

double nuIndEst(double S, double r, double dividend, double sigma,
		double T, double t, double strikePrice);

void option_error(char *argv[]){
  printf("Usage: %s parameters\n",argv[0]);
  printf("\t-n underliers (n from Broadie and Glasserman)\n");
  printf("\t-r risk-free rate (r)\n");
  printf("\t-D dividend rate (\\delta)\n");
  printf("\t-v variance (\\sigmna^2)\n");
  printf("\t-T time to maturity (T)\n");
  printf("\t-K strikeprice (K)\n");
  printf("\t-S initial price (S_0)\n");
  printf("\t-N repetitions (N)\n");
  printf("\t-b paths in mesh (b)\n");
  printf("\t-p paths in path estimate (n_p)\n");
  printf("\t-d time slices (d)\n");
  printf("\t-f number of standard deviations kept in filtered average\n");
  printf("\t-i which inner control (1-3)\n");
  printf("\t-c number of paths in a cluster\n");
  printf("\t-m cluster variance\n");
  printf("\nRates and variance are for T=1.\n");
  printf("All parameters are required.\n\n");

  exit(1);
}

int main(int argc, char *argv[]){
  meshData md;
  int i,j,time, path, underlier, mac, pac, cmac, cpac;
  double europe_value, *me,*pe,*cme,*cpe, ma, mv, pa, pv, cma, cmv, cpa, cpv, nma, npa, nmv, npv, ncma, ncmv, ncpa, ncpv;
  int reps, reps_bar;
/* variables need for OuterCtl routine */
	double *option_t, *option_tp1;
	double *d_european, *rho_european, **rho2d_european, uhat_avg;
	double Qbar_var, Qbar_avg;

  const char *options = "n:r:D:v:T:K:S:N:b:p:d:f:m:c:i:";
  int n,N,N_bar,b,p,d,option,count=0,c,whichControl;
  double T,r,D,v,K,S,f,m;

   /* Arrays for Outer controls */
   double *Qhat, **uhat, *u;
   double Qbar;

  /* Read command line parameters */
  while((option = getopt(argc,argv,options)) != EOF){
    switch(option){
    case('n'):
      sscanf(optarg,"%i",&n);
      count++;
      break;
    case('r'):
      sscanf(optarg,"%lf",&r);
      count++;
      break;
    case('D'):
      sscanf(optarg,"%lf",&D);
      count++;
      break;
    case('v'):
      sscanf(optarg,"%lf",&v);
      count++;
      break;
    case('T'):
      sscanf(optarg,"%lf",&T);
      count++;
      break;
    case('K'):
      sscanf(optarg,"%lf",&K);
      count++;
      break;
    case('S'):
      sscanf(optarg,"%lf",&S);
      count++;
      break;
    case('N'):
      sscanf(optarg,"%i",&N);
      count++;
      break;
    case('i'):
      sscanf(optarg,"%i",&whichControl);
      count++;
      break;
    case('B'):
      sscanf(optarg,"%i",&N_bar);
      count++;
      break;
    case('b'):
      sscanf(optarg,"%i",&b);
      count++;
      break;
    case('p'):
      sscanf(optarg,"%i",&p);
      count++;
      break;
    case('d'):
      sscanf(optarg,"%i",&d);
      count++;
      break;
    case('c'):
      sscanf(optarg,"%i",&c);
      count++;
      break;
    case('f'):
      sscanf(optarg,"%lf",&f);
      count++;
      break;
    case('m'):
      sscanf(optarg,"%lf",&m);
      count++;
      break;
    default:
      option_error(argv);
    }
  }

  printf("%i\n",count);
  if(count != 15){
    option_error(argv);
  }

  printf("n = %i\n",n);
  printf("T = %lf\n",T);
  printf("N = %i\n",N);
  printf("N_bar = %i\n",N_bar);
  printf("b = %i\n",b);
  printf("p = %i\n",p);
  printf("d = %i\n",d);
  printf("r = %lf\n",r);
  printf("D = %lf\n",D);
  printf("v = %lf\n",v);
  printf("K = %lf\n",K);
  printf("S = %lf\n",S);
  printf("c = %i\n",c);
  printf("f = %lf\n",f);
  printf("m = %lf\n",m);

  allocate1dDouble(&me,N);
  allocate1dDouble(&pe,N);
  allocate1dDouble(&cme,N);
  allocate1dDouble(&cpe,N);

  /* Allocate memory for Outer controls */
   allocate1dDouble(&Qhat, N);
   allocate2dDouble(&uhat, 2,N);
   allocate1dDouble(&u, 2);
   allocate1dDouble(&option_t,   b);
   allocate1dDouble(&option_tp1, b);
   allocate1dDouble(&d_european, n);
   allocate1dDouble(&rho_european, n*(n-1)/2);
   allocate2dDouble(&rho2d_european, n, n);

  /* The parameters that can be set before memory allocation */
  /* times, paths and underliers required to come before allocation. */
  /* Later, these should go in a config file. */
  md.option.pData.times = d + 1;
  md.option.pData.paths = b + p;
  md.option.pData.underliers = n;
  md.option.pData.clusterSize = c;
  md.option.pData.clusterVariance = m;
  md.option.strikePrice = K;
  md.option.riskFreeRate = r*(double)T/(double)d;
  md.meshPaths = b;
  md.option.pData.meshPaths = b;
  md.tolerance = 1.0;
  strcpy(md.option.pData.phrase,PHRASE);

  /* Allocate the memory */
  pathDataAllocate(&(md.option.pData));
  optionDataAllocate(&(md.option));
  meshDataAllocate(&md);

  /* Fill the covariance matrix and mean vector */
  /* In our case, the covariance matrix is diagonal with the variance */
  /* in all diagonal entries. */
  for(i=0;i<md.option.pData.underliers;i++){
    for(j=0;j<md.option.pData.underliers;j++){
      if(i == j){
	md.option.pData.covariance[i][j] = v*(double)T/(double)d;
	md.option.pData.growthRate[i] = r*(double)T/(double)d;
	md.option.pData.dividendRate[i] = D*(double)T/(double)d;
      }else{
	md.option.pData.covariance[i][j] = 0;
      }
    }
  }


/*  N_bar = N; */
  /* Sets up internal RNG stuff */
  /* Must come after the means and covariances */
  pathParamsInit(&(md.option.pData));

	Qbar_avg = 0;
  /* Set all the initial prices */
  for(path=0; path<md.option.pData.paths; path++){
    for(underlier=0; underlier < md.option.pData.underliers; underlier++){
      md.option.pData.price[0][path][underlier] = S;
    }
  }

  ma = 0;
  pa = 0;
  mv = 0;
  pv = 0;
  cma  = 0;
  cpa = 0;
  cmv = 0;
  cpv = 0;
	uhat_avg = 0;
  for(reps=0;reps<N;reps++){
    pathDataFill(&(md.option.pData));
    optionDataFill(&(md.option));
    meshSetFinalValue(&md);
    meshDataFill(&md); 

    me[reps] = md.valueIfHeld[0][0];
    pe[reps] = meshPathEstimate(&md);

    ma += me[reps];
    mv += me[reps]*me[reps];
    pa += pe[reps];
    pv += pe[reps]*pe[reps];

    innerDataFill(&md,whichControl);

    cme[reps] = md.newValueIfHeld[0][0];
    cpe[reps] = innerControlledPathEstimate(&md);

    cma += cme[reps];
    cmv += cme[reps]*cme[reps];
    cpa += cpe[reps];
    cpv += cpe[reps]*cpe[reps];

    printf("%lf\t%lf\t%lf\t%lf\n",me[reps],pe[reps],cme[reps],cpe[reps]);

    
    /* compute mesh estimates of u1 and/or u2 */ 
		Qhat[reps] = cme[reps];
		uhat[0][reps] = OuterCtlEuroMaxMeshValue(&md,  r,  D, (double)T,
                                               (double)(T/d),option_t,
                                               option_tp1);
		uhat[1][reps] = OuterCtlEuroMaxMeshValue(&md,  r,  D, 2.0/3.0*(double)T,
                                               (double)(T/d),option_t,
                                               option_tp1);
		printf("Main>>> uhat[1][%d]=%lf  uhat[2][%d]=%lf Qhat[%d]=%lf\n",
                reps,uhat[0][reps],reps,uhat[1][reps],reps,Qhat[reps]); 
  }
	for (reps=0;reps<N;reps++)
	{
		uhat_avg += uhat[0][reps];
	}
	printf("Main>>> uhat1_avg=%lf\n",uhat_avg/N);

	for (reps=0;reps<N;reps++)
	{
		uhat_avg += uhat[1][reps];
	}
	printf("Main>>> uhat2_avg=%lf\n",uhat_avg/N);
  /* compute independent value of u1 and/or u2 */
   u[0] = EuropeanMaxOption(&md, r, v, D, T, d_european, rho_european,
                            rho2d_european);
   u[1] = EuropeanMaxOption(&md, r, v, D, 2.0*(double)(T)/3.0,
                            d_european, rho_european, rho2d_european);
	printf("Main>>> u[0]=%lf u[1]=%lf\n",u[0],u[1]);
   switch((int)S){
    case(90):
      u[0] = 14.586;
      u[1] = 0.0;
    break;
    case(100):
      u[0] = 23.052;
      u[1] = 0.0;
    break;
    case(110):
      u[0] = 32.685;
      u[1] = 0.0;
    break;
    default:
      perror("No European value present for this initial price");
      exit(1);
   }
	printf("Main>>> u[0]=%lf u[1]=%lf\n",u[0],u[1]);

  /* Compute outer control estimate  */
  Qbar = Compute_Outer_Control_Estimate(1, N, Qhat, uhat, u);
/* The N indivual Qhats have been updated in the above routine with Qbar 
   being the average of the updated Qhats. */
	Qbar_var = 0.0;
	for (i=0; i<N; i++)
	{
		Qbar_var += Qhat[i]*Qhat[i];
	}
	Qbar_var = sqrt(Qbar_var/N - Qbar*Qbar);

  ma /= N;
  mv = mv/N - ma*ma;
  pa /= N;
  pv = pv/N - pa*pa;
  cma /= N;
  cmv = cmv/N - cma*cma;
  cpa /= N;
  cpv = cpv/N - cpa*cpa;
  printf("Values with inner controls applied \n");
  printf("Est: %lf\t%lf\t%lf\t%lf\n",ma,pa,cma,cpa);
  printf("SDs: %lf\t%lf\t%lf\t%lf\n",sqrt(mv),sqrt(pv),sqrt(cmv),sqrt(cpv));
  printf("SDMs: %lf\t%lf\t%lf\t%lf\n",sqrt(mv/N),sqrt(pv/N),sqrt(cmv/N),sqrt(cpv/N));

    printf("\n\n After outer control 1\n");
	printf("Qbar=%lf\n",Qbar);
	printf("Qbar variance = %lf std dev = %lf \n",Qbar_var,sqrt(Qbar_var));

  nma = 0;
  npa = 0;
  nmv = 0;
  npv = 0;
  ncma  = 0;
  ncpa = 0;
  ncmv = 0;
  ncpv = 0;  
  mac = 0;
  pac = 0;
  cmac = 0;
  cpac = 0;
  for(reps=0;reps<N;reps++){
    if(fabs(me[reps] - ma) < f*sqrt(mv)){
      nma += me[reps];
      nmv += me[reps]*me[reps];
      mac++;
    }
    if(fabs(pe[reps] - pa) < f*sqrt(pv)){
      npa += pe[reps];
      npv += pe[reps]*pe[reps];
      pac++;
    }
    if(fabs(cme[reps] - cma) < f*sqrt(cmv)){
      ncma += cme[reps];
      ncmv += cme[reps]*cme[reps];
      cmac++;
    }
    if(fabs(cpe[reps] - cpa) < f*sqrt(cpv)){
      ncpa += cpe[reps];
      ncpv += cpe[reps]*cpe[reps];
      cpac++;
    }
  }

  printf("\n");
  nma /= mac;
  npa /= pac;
  nmv = nmv/mac - nma*nma;
  npv = npv/pac - npa*npa;
  ncma /= cmac;
  ncpa /= cpac;
  ncmv = ncmv/cmac - ncma*ncma;
  ncpv = ncpv/cpac - ncpa*ncpa;  
  printf("%lf\t%lf\t%lf\t%lf\n",nma,npa,ncma,ncpa);
  printf("%lf\t%lf\t%lf\t%lf\n",sqrt(nmv),sqrt(npv),sqrt(ncmv),sqrt(ncpv));
  printf("%lf\t%lf\t%lf\t%lf\n",sqrt(nmv/N),sqrt(npv/N),sqrt(ncmv/N),sqrt(ncpv/N));





}

void sumWeights(meshData *meshData){
  int t, i, j;
  double sum;
  FILE *fp;

  fp = fopen("wghtavgs.dat", "w+");

  for(t=0; t<meshData->option.pData.times; t++){
    sum = 0.0;
    for(i=0; i < meshData->option.pData.paths; i++){
      for(j=0; j < meshData->meshPaths; j++){
        sum += meshData->weights[t][i][j];
      }
    }
    fprintf(fp,"t = %d  sum = %lf \n",t,sum); 
  }
  fclose(fp);

}
